#This is a simple file for storing the automatically loaded trace_styles_collection dictionary for JSONGrapher.
#It is not recommended that a person overwrite this file.

#each dictionary below is one 'trace_styles_collection'. "default" is the first one.
styles_library = {
    "default": {
        "scatter_spline": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 10},
        },
        "scatter_line": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "linear", "width": 2},
            "marker": {"size": 10},
        },
        "line": {
            "type": "scatter",
            "mode": "lines",
            "line": {"shape": "linear", "width": 2},
            "marker": {"size": 10},
        },
        "lines": {
            "type": "scatter",
            "mode": "lines",
            "line": {"shape": "linear", "width": 2},
            "marker": {"size": 10},
        },
        "scatter": {
            "type": "scatter",
            "mode": "markers",
            "marker": {"size": 10},
        },
        "bubble": {
            "type": "scatter",
            "mode": "markers",
            "marker": {
                            "color": "auto",
                            "colorscale": "viridis_r",
                            "showscale": True
                        }
        },        
        "bubble2d": {
            "type": "scatter",
            "mode": "markers",
            "marker": {
                            "color": "auto",
                            "colorscale": "viridis_r",
                            "showscale": True
                        }
        },     
        "spline": {
            "type": "scatter",
            "mode": "lines",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 0},  # Hide markers for smooth curves
        },
        "bar": {
            "type": "bar",
            "marker": {
                            "color": "blue",
                            "opacity": 0.8,
                            "line": {
                                "color": "black",
                                "width": 2
                            }
                        },
        },
        "default": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 10},
        },        
        "scatter3d": {
            "mode": "markers",
            "type": "scatter3d",
            "marker": {"color" : "","colorscale":"rainbow", "showscale":True}
        },
        "mesh3d": {
            "type": "mesh3d",
            "intensity" : [],
            "colorscale":"rainbow", 
            "showscale":True
        },
        "bubble3d": {
            "mode": "markers",
            "type": "scatter3d",
            "marker": {"color" : "","colorscale":"rainbow", "showscale":True}
        },
        "heatmap": {
            "type": "heatmap",
            "colorscale": "Viridis",
        }
    },
    "minimalist": {
        "scatter_spline": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 1},
            "marker": {"size": 6},
        },
        "scatter": {
            "type": "scatter",
            "mode": "lines",
            "line": {"shape": "linear", "width": 1},
            "marker": {"size": 0},
        },
        "spline": {
            "type": "scatter",
            "mode": "lines",
            "line": {"shape": "spline", "width": 1},
            "marker": {"size": 0},
        },
        "bar": {
            "type": "bar",
        },
        "heatmap": {
            "type": "heatmap",
            "colorscale": "Greys",
        }
    },
    "bold": {
        "scatter_spline": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 4},
            "marker": {"size": 10},
        },
        "scatter": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 4},
            "marker": {"size": 12},
        },
        "spline": {
            "type": "scatter",
            "mode": "lines",
            "line": {"shape": "spline", "width": 4},
            "marker": {"size": 0},
        },
        "bar": {
            "type": "bar",
        },
        "heatmap": {
            "type": "heatmap",
            "colorscale": "Jet",
        }
    },
    "scatter": { #this style forces all traces into scatter.
        "scatter_spline": {
            "type": "scatter",
            "mode": "markers",
            "marker": {"size": 10},
        },
        "scatter": {
            "type": "scatter",
            "mode": "markers",
            "marker": {"size": 10},
        },
        "spline": {
            "type": "scatter",
            "mode": "markers",
            "marker": {"size": 10},
        },
        "bar": {
            "type": "scatter",
            "mode": "markers",
            "marker": {"size": 10},
        },
        "heatmap": {
            "type": "heatmap",
            "colorscale": "Viridis",
        }
    },
    "scatter_spline": { #this style forces all traces into spline only
        "scatter_spline": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 0},
        },
        "scatter": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 0},
        },
        "spline": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 0},
        },
        "bar": {
            "type": "scatter",
            "mode": "lines+markers",
            "line": {"shape": "spline", "width": 2},
            "marker": {"size": 0},
        },
        "heatmap": {
            "type": "heatmap",
            "colorscale": "Viridis",
        }
    }
}


import unitpy
import re
from unitpy.definitions.entry import Entry

def add_custom_unit(unit_string):
    # Need to put an entry into "bases" because the BaseSet class will pull from that dictionary.
    unitpy.definitions.unit_base.bases[unit_string] = unitpy.definitions.unit_base.BaseUnit(
        label=unit_string, abbr=unit_string, dimension=unitpy.definitions.dimensions.dimensions["amount_of_substance"]
    )
    
    # Then need to make a BaseSet object to put in. Confusingly, we *do not* put a BaseUnit object into the base_unit argument, below. 
    # We use "mole" to avoid conflicting with any other existing units.
    base_unit = unitpy.definitions.unit_base.BaseSet(mole=1)
    
    new_entry = Entry(label=unit_string, abbr=unit_string, base_unit=base_unit, multiplier=1)
    
    # Only add the entry if it is missing. A duplicate entry would cause crashing later.
    if 'frog' not in unitpy.ledger._lookup:
        unitpy.ledger.add_unit(new_entry)  # Implied return is here. No return needed.

add_custom_unit("frog")
add_custom_unit("frog")

units_string_1 = 'm*frog'
units_string_2 = 'm*frog'
units_string_1_multiplied = 1*unitpy.U(units_string_1 )
print("line 25", type(units_string_1_multiplied))
units_string_1_multiplied.to(units_string_2)



import re
import json

try:
    from json_equationer.equation_evaluator import evaluate_equation_dict
except ImportError:
    try:
        from .equation_evaluator import evaluate_equation_dict
    except ImportError:
        from equation_evaluator import evaluate_equation_dict


class Equation:
    """
    A class to manage mathematical equations with units and to evaluate them.
    Provides utilities for evaluating, formatting, exporting, and printing.
   
    Initialization:
    - Normally, should be initialized as a blank dict object like example_Arrhenius = Equation().
    - Defaults to an empty equation with predefined structure.
    - Accepts an optional dictionary (`initial_dict`) to prepopulate the equation dictionary.

    Example structure:
    ```
    custom_dict = {
        'equation_string': "k = A * (e ** (-Ea / (R * T)))",
        'x_variable': "T (K)",
        'y_variable': "k (s**-1)",
        'constants': {"Ea": "30000 J/mol", "R": "8.314 J/(mol*K)", "A": "1*10**13 (s**-1)", "e": "2.71828"},
        'num_of_points': 10,
        'x_range_default': [200, 500],
        'x_range_limits': [None, 600],
        'points_spacing': "Linear"
        'graphical_dimensionality': 2
    }

    #The reason we use 'graphical_dimensionality' rather than 'dimensionality' is that mathematicians define the dimensionality in terms of independent variables.
    #But here, we are usually expecting users who are concerned with 2D or 3D graphing.

    equation_instance = Equation(initial_dict=custom_dict)
    ```
    """

    def __init__(self, initial_dict=None):
        """Initialize an empty equation dictionary."""
        if initial_dict==None:
            initial_dict = {}
        self.equation_dict = {
            'equation_string': '',
            'x_variable': '',  
            'y_variable': '',
            'constants': {},
            'num_of_points': None,  # Expected: Integer, defines the minimum number of points to be calculated for the range.
            'x_range_default': [0, 1],  # Default to [0,1] instead of an empty list.
            'x_range_limits': [None, None],  # Allows None for either limit.
            'x_points_specified': [],
            'points_spacing': '',
            'reverse_scaling': False,
        }

        # If a dictionary is provided, update the default values
        if len(initial_dict)>0:
            if isinstance(initial_dict, dict):
                self.equation_dict.update(initial_dict)
            else:
                raise TypeError("initial_dict must be a dictionary.")

    def validate_unit(self, value):
        """Ensure that the value is either a pure number or contains a unit."""
        unit_pattern = re.compile(r"^\d+(\.\d+)?(.*)?$")
        if not unit_pattern.match(value):
            raise ValueError(f"Invalid format: '{value}'. Expected a numeric value, optionally followed by a unit.")

    def add_constants(self, constants):
        """Add constants to the equation dictionary, supporting both single and multiple additions."""
        if isinstance(constants, dict):  # Single constant case
            for name, value in constants.items():
                self.validate_unit(value)
                self.equation_dict['constants'][name] = value
        elif isinstance(constants, list):  # Multiple constants case
            for constant_dict in constants:
                if isinstance(constant_dict, dict):
                    for name, value in constant_dict.items():
                        self.validate_unit(value)
                        self.equation_dict['constants'][name] = value
                else:
                    raise ValueError("Each item in the list must be a dictionary containing a constant name-value pair.")
        else:
            raise TypeError("Expected a dictionary for one constant or a list of dictionaries for multiple constants.")

    def set_x_variable(self, x_variable):
        """
        Set the x-variable in the equation dictionary.
        Expected format: A descriptive string including the variable name and its unit.
        Example: "T (K)" for temperature in Kelvin.
        """
        self.equation_dict["x_variable"] = x_variable

    def set_y_variable(self, y_variable):
        """
        Set the y-variable in the equation dictionary.
        Expected format: A descriptive string including the variable name and its unit.
        Example: "k (s**-1)" for a rate constant with inverse seconds as the unit.
        """
        self.equation_dict["y_variable"] = y_variable

    def set_z_variable(self, z_variable):
        """
        Set the z-variable in the equation dictionary.
        Expected format: A descriptive string including the variable name and its unit.
        Example: "E (J)" for energy with joules as the unit.
        """
        self.equation_dict["z_variable"] = z_variable

    def set_x_range_default(self, x_range):
        """
        Set the default x range.
        Expected format: A list of two numeric values representing the range boundaries.
        Example: set_x_range([200, 500]) for temperatures between 200K and 500K.
        """
        if not (isinstance(x_range, list) and len(x_range) == 2 and all(isinstance(i, (int, float)) for i in x_range)):
            raise ValueError("x_range must be a list of two numeric values.")
        self.equation_dict['x_range_default'] = x_range

    def set_x_range_limits(self, x_limits):
        """
        Set the hard limits for x values.
        Expected format: A list of two values (numeric or None) defining absolute boundaries.
        Example: set_x_range_limits([100, 600]) to prevent x values outside this range.
        Example: set_x_range_limits([None, 500]) allows an open lower limit.
        """
        if not (isinstance(x_limits, list) and len(x_limits) == 2):
            raise ValueError("x_limits must be a list of two elements (numeric or None).")
        if not all(isinstance(i, (int, float)) or i is None for i in x_limits):
            raise ValueError("Elements in x_limits must be numeric or None.")
        self.equation_dict['x_range_limits'] = x_limits

    def set_y_range_default(self, y_range):
        """
        Set the default y range.
        Expected format: A list of two numeric values representing the range boundaries.
        Example: set_y_range([0, 100]) for a percentage scale.
        """
        if not (isinstance(y_range, list) and len(y_range) == 2 and all(isinstance(i, (int, float)) for i in y_range)):
            raise ValueError("y_range must be a list of two numeric values.")
        self.equation_dict['y_range_default'] = y_range

    def set_y_range_limits(self, y_limits):
        """
        Set the hard limits for y values.
        Expected format: A list of two values (numeric or None) defining absolute boundaries.
        Example: set_y_range_limits([None, 50]) allows an open lower limit but restricts the upper limit.
        """
        if not (isinstance(y_limits, list) and len(y_limits) == 2):
            raise ValueError("y_limits must be a list of two elements (numeric or None).")
        if not all(isinstance(i, (int, float)) or i is None for i in y_limits):
            raise ValueError("Elements in y_limits must be numeric or None.")
        self.equation_dict['y_range_limits'] = y_limits

    def set_z_range_default(self, z_range):
        """
        Set the default z range.
        Expected format: A list of two numeric values representing the range boundaries.
        Example: set_z_range([0, 5000]) for energy values in Joules.
        """
        if not (isinstance(z_range, list) and len(z_range) == 2 and all(isinstance(i, (int, float)) for i in z_range)):
            raise ValueError("z_range must be a list of two numeric values.")
        self.equation_dict['z_range_default'] = z_range

    def set_z_range_limits(self, z_limits):
        """
        Set the hard limits for z values.
        Expected format: A list of two values (numeric or None) defining absolute boundaries.
        Example: set_z_range_limits([100, None]) allows an open upper limit but restricts the lower boundary.
        """
        if not (isinstance(z_limits, list) and len(z_limits) == 2):
            raise ValueError("z_limits must be a list of two elements (numeric or None).")
        if not all(isinstance(i, (int, float)) or i is None for i in z_limits):
            raise ValueError("Elements in z_limits must be numeric or None.")
        self.equation_dict['z_range_limits'] = z_limits

    def get_z_matrix(self, x_points=None, y_points=None, z_points=None, return_as_list=False):
        """
        Constructs a Z matrix mapping unique (x, y) values to corresponding z values.

        Parameters:
        - x_points (list): List of x coordinates.
        - y_points (list): List of y coordinates.
        - z_points (list): List of z values.
        - return_as_list (bool, optional): Whether to return the matrix as a list. Defaults to False (returns NumPy array).

        Returns:
        - z_matrix (2D list or numpy array): Matrix of z values.
        - unique_x (list): Sorted unique x values.
        - unique_y (list): Sorted unique y values.
        """
        if x_points == None:
            x_points = self.equation_dict['x_points']
        if y_points == None:
            y_points = self.equation_dict['y_points']
        if z_points == None:
            z_points = self.equation_dict['z_points']

        import numpy as np
        # Get unique x and y values
        unique_x = sorted(set(x_points))
        unique_y = sorted(set(y_points))

        # Create an empty matrix filled with NaNs
        z_matrix = np.full((len(unique_x), len(unique_y)), np.nan)

        # Map z values to corresponding x, y indices
        for x, y, z in zip(x_points, y_points, z_points):
            x_idx = unique_x.index(x)
            y_idx = unique_y.index(y)
            z_matrix[x_idx, y_idx] = z

        # Convert to a list if requested
        if return_as_list:
            z_matrix = z_matrix.tolist()

        return z_matrix

    



    def set_num_of_points(self, num_points):
        """
        Set the number of calculation points.
        Expected format: Integer, specifies the number of discrete points for calculations.
        Example: set_num_of_points(10) for ten data points.
        """
        if not isinstance(num_points, int) or num_points <= 0:
            raise ValueError("Number of points must be a positive integer.")
        self.equation_dict["num_of_points"] = num_points

    def set_equation(self, equation_string):
        """Modify the equation string."""
        self.equation_dict['equation_string'] = equation_string

    def get_equation_dict(self):
        """Return the complete equation dictionary."""
        return self.equation_dict
    
    def evaluate_equation(self, remove_equation_fields= False, verbose=False):
        evaluated_dict = evaluate_equation_dict(self.equation_dict, verbose=verbose) #this function is from the evaluator module
        if "graphical_dimensionality" in evaluated_dict:
            graphical_dimensionality = evaluated_dict["graphical_dimensionality"]
        else:
            graphical_dimensionality = 2
        self.equation_dict["x_units"] = evaluated_dict["x_units"]
        self.equation_dict["y_units"] = evaluated_dict["y_units"]
        self.equation_dict["x_points"] = evaluated_dict["x_points"]
        self.equation_dict["y_points"] = evaluated_dict["y_points"]
        if graphical_dimensionality == 3:
            self.equation_dict["z_points"] = evaluated_dict["z_points"]
        if remove_equation_fields == True:
            #we'll just make a fresh dictionary for simplicity, in this case.
            equation_dict = {}
            equation_dict["x_units"] = self.equation_dict["x_units"] 
            equation_dict["y_units"] = self.equation_dict["y_units"]
            equation_dict["x_points"] = self.equation_dict["x_points"] 
            equation_dict["y_points"] = self.equation_dict["y_points"] 
            if graphical_dimensionality == 3:
                equation_dict["z_units"] = self.equation_dict["z_units"]
                equation_dict["z_points"] = self.equation_dict["z_points"] 
                print("line 223", equation_dict["z_points"])
            self.equation_dict = equation_dict
        return self.equation_dict

    def print_equation_dict(self, pretty_print=True, evaluate_equation = True, remove_equation_fields = False):
        equation_dict = self.equation_dict #populate a variable internal to this function.
        #if evaluate_equation is true, we'll try to simulate any series that need it, then clean the simulate fields out if requested.
        if evaluate_equation == True:
            evaluated_dict = self.evaluate_equation(remove_equation_fields = remove_equation_fields) #For this function, we don't want to remove equation fields from the object, just the export.
            equation_dict = evaluated_dict
        if remove_equation_fields == True:
            equation_dict = {}
            equation_dict["x_units"] = self.equation_dict["x_units"] 
            equation_dict["y_units"] = self.equation_dict["y_units"]
            equation_dict["x_points"] = self.equation_dict["x_points"] 
            equation_dict["y_points"] = self.equation_dict["y_points"] 
        if pretty_print == False:
            print(equation_dict)
        if pretty_print == True:
            equation_json_string = json.dumps(equation_dict, indent=4)
            print(equation_json_string)

    def export_to_json_file(self, filename, evaluate_equation = True, remove_equation_fields= False):
        """
        writes the json to a file
        returns the json as a dictionary.
        update_and_validate function will clean for plotly. One can alternatively only validate.
        optionally simulates all series that have a simulate field (does so by default)
        optionally removes simulate filed from all series that have a simulate field (does not do so by default)
        optionally removes hints before export and return.
        """
        equation_dict = self.equation_dict #populate a variable internal to this function.
        #if evaluate_equation is true, we'll try to simulate any series that need it, then clean the simulate fields out if requested.
        if evaluate_equation == True:
            evaluated_dict = self.evaluate_equation(remove_equation_fields = remove_equation_fields) #For this function, we don't want to remove equation fields from the object, just the export.
            equation_dict = evaluated_dict
        if remove_equation_fields == True:
            equation_dict = {}
            equation_dict["x_units"] = self.equation_dict["x_units"] 
            equation_dict["y_units"] = self.equation_dict["y_units"]
            equation_dict["x_points"] = self.equation_dict["x_points"] 
            equation_dict["y_points"] = self.equation_dict["y_points"] 
        # filepath: Optional, filename with path to save the JSON file.       
        if len(filename) > 0: #this means we will be writing to file.
            # Check if the filename has an extension and append `.json` if not
            if '.json' not in filename.lower():
                filename += ".json"
            #Write to file using UTF-8 encoding.
            with open(filename, 'w', encoding='utf-8') as f:
                json.dump(equation_dict, f, indent=4)
        return equation_dict



if __name__ == "__main__":
    # Create an instance of Equation
    example_Arrhenius = Equation()
    example_Arrhenius.set_equation("k = A * (e ** (-Ea / (R * T)))")
    example_Arrhenius.set_x_variable("T (K)")  # Temperature in Kelvin
    example_Arrhenius.set_y_variable("k (s**-1)")  # Rate constant in inverse seconds

    # Add a constants one at a time, or through a list.
    example_Arrhenius.add_constants({"Ea": "30000 J/mol"})  
    example_Arrhenius.add_constants([
        {"R": "8.314 J/(mol*K)"},
        {"A": "1*10**13 (s**-1)"},
        {"e": "2.71828"}  # No unit required
    ])

    # Optinally, set minimum number of points and limits for calculations.
    example_Arrhenius.set_num_of_points(10)
    example_Arrhenius.set_x_range_default([200, 500])
    example_Arrhenius.set_x_range_limits([None, 600])  

    # Define additional properties.
    example_Arrhenius.equation_dict["points_spacing"] = "Linear"

    # Retrieve and display the equation dictionary
    example_equation_dict = example_Arrhenius.get_equation_dict()
    print(example_equation_dict)

    example_Arrhenius.evaluate_equation()
    example_Arrhenius.print_equation_dict()


    #Now for a 3D example.
    example_Arrhenius_3D_dict = {
        'equation_string': 'k = A*(e**((-Ea)/(R*T)))',
        'graphical_dimensionality' : 3,
        'x_variable': 'T (K)',  
        'y_variable': 'Ea (J)*(mol^(-1))',
        'z_variable': 'k (s**(-1))', 
        'constants': {'R': '8.314 (J)*(mol^(-1))*(K^(-1))' , 'A': '1*10^13 (s^-1)', 'e': '2.71828'},
        'num_of_points': 10,
        'x_range_default': [200, 500],
        'x_range_limits' : [],
        'y_range_default': [30000, 50000],
        'y_range_limits' : [],
        'x_points_specified' : [],
        'points_spacing': 'Linear',
        'reverse_scaling' : False
    }

    example_Arrhenius_3D_equation = Equation(initial_dict=example_Arrhenius_3D_dict)
    evaluated_output = example_Arrhenius_3D_equation.evaluate_equation()
    #print(evaluated_output)
    #print(example_Arrhenius_3D_equation.get_z_matrix(return_as_list=True))



#Try to do things a bit like https://github.com/python/cpython/blob/master/Lib/collections/__init__.py
#Except instead of putting things here directly in __init__, we'll import them so they are accessible by importing the module.
from JSONGrapher.JSONRecordCreator import *
from JSONGrapher.version import *


#This is a simple file for storing the automatically loaded trace_styles_collection dictionary for JSONGrapher.
#It is not recommended that a person overwrite this file.

#each dictionary below, like "default" or "Nature" is one 'layout_style'.
#Currently, the style names are case sensitive. In the future, it is likely they will not be case sensitive.
styles_library = {
        "default": {
            "layout": {
                "title": {"font": {"size": 32}, "x": 0.5},
                "xaxis": {"title": {"font": {"size": 27}}, "tickfont": {"size": 23}},
                "yaxis": {"title": {"font": {"size": 27}}, "tickfont": {"size": 23}},
                "legend": {
                    "title": {"font": {"size": 22}},
                    "font": {"size": 22}
                }
            }
        },
        "default3d": {
            "layout": {
                "scene": {"aspectmode":"cube"},
                "title": {"font": {"size": 32}, "x": 0.5},
                "xaxis": {"title": {"font": {"size": 12}}, "tickfont": {"size": 12}},
                "yaxis": {"title": {"font": {"size": 12}}, "tickfont": {"size": 12}},
                "zaxis": {"title": {"font": {"size": 12}}, "tickfont": {"size": 12}},
                "legend": {
                    "title": {"font": {"size": 22}},
                    "font": {"size": 22}
                }
            }
        },
        "Nature": {
            "layout": {
                "title": {"font": {"size": 32, "family": "Times New Roman", "color": "black"}},
                "font": {"size": 25, "family": "Times New Roman"},
                "paper_bgcolor": "white",
                "plot_bgcolor": "white",
                "xaxis": {
                    "showgrid": True, "gridcolor": "#ddd", "gridwidth": 1,
                    "linecolor": "black", "linewidth": 2, "ticks": "outside",
                    "tickwidth": 2, "tickcolor": "black"
                },
                "yaxis": {
                    "showgrid": True, "gridcolor": "#ddd", "gridwidth": 1,
                    "linecolor": "black", "linewidth": 2, "ticks": "outside",
                    "tickwidth": 2, "tickcolor": "black"
                }
            }
        },
        "Science": {
            "layout": {
                "title": {"font": {"size": 32, "family": "Arial", "color": "black"}},
                "font": {"size": 25, "family": "Arial"},
                "paper_bgcolor": "white",
                "plot_bgcolor": "white",
                "xaxis": {
                    "showgrid": True, "gridcolor": "#ccc", "gridwidth": 1,
                    "linecolor": "black", "linewidth": 2, "ticks": "outside",
                    "tickwidth": 2, "tickcolor": "black"
                },
                "yaxis": {
                    "showgrid": True, "gridcolor": "#ccc", "gridwidth": 1,
                    "linecolor": "black", "linewidth": 2, "ticks": "outside",
                    "tickwidth": 2, "tickcolor": "black"
                }
            }
        }
    }






'''
    #This units list came from the UUC units list. https://github.com/Lemonexe/UUC/blob/master/app/data.js
    # it has been converted to python and contains all of the program constants, the unit database, and database of prefixes
    #In the python JSONGrapher repository, this is used (as of May 2025) simply to remove "plural" units.
    # In python JSONGrapher, the unitpy package is what is used for unit conversion.
    # 'en' is for English, 'cz' is for Czech, and "ae" is for American English.

    metre → meter
    mole → mol (sometimes abbreviated in scientific contexts)
    dioptre → diopter
    normal cubic metre → normal cubic meter
    normal litre → normal liter
    milimetre of mercury → millimeter of mercury
    litre → liter
    tonne → metric ton (often written as ton in American English)
    cubic centimetre → cubic centimeter

This file is adapted from UUC and is covered by the MIT License
        
MIT License

Copyright (c) 2017 Jiri Zbytovsky

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in all
copies or substantial portions of the Software.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
SOFTWARE.


'''
import math as Math
#program constants
dict_csts = {
    "R": 8.3144598,  # [J/K/mol]
    "TC0": 273.15,  # [K]
    "TF0": 459.67 * 5 / 9,  # [K]
    "atm": 101325,  # [Pa]
    "mile": 1609.344,  # [m]
    "bbl": 158.987294928e-3,  # [m3]
    "q": 1.6021766208e-19,  # [C]
    "BTU": 1055.05585,  # [J]
    "APIk": 141.5e3,  # [1]
    "APIq": -131.5,  # [m3/kg]
    "dTnote": {"cz": "Viz °C pro důležitou poznámku.", "en": "See °C for an important note."},
    "Hgnote": {"cz": "Mezi mmHg a Torr je nepatrný rozdíl.", "en": "There is a negligible difference between mmHg and Torr."}
}

#Class to access the above dictionary objects with notation like javascript.
class Csts:
    def __init__(self, constants):
        self.__dict__.update(constants)

# Create an instance of Csts
csts = Csts(dict_csts)


'''
Units object is the database of all known units.
    v: [m,kg,s,A,K,mol,cd,$]      represents the vector of powers of basic units, for example N=kg*m/s^2, therefore v = [1,1,-2,0,0,0,0]
    id: string                    something unique. You can use the UnitConflicts() global function to detect possible id conflicts
    alias: array                  array of strings - other ids that reference the unitprefix:   
    name: object                  defines full name or even a short description for every language mutation
    k: number                     this coeficient equates value of the unit in basic units. For example minute = 60 seconds, therefore min.k = 60
    "SI": True/false                self-explanatory. This attribute doesn't really do anything, it's merely informational. Perhaps it's redundant, since all SI derived units have k = 1
    basic: True/false             whether it's basic SI unit or derived SI. Basic SI units are of utmost importance to the code, don't ever change them!
    prefix: all/+/-/undefined     it means: all prefixes allowed / only bigger than one allowed / lesser than one / prefixes disallowed. It's not really a restriction, just a recommendation.
    "constant": True/undefined      whether it is a constant. If true, attributes SI, basic and prefix are ignored. Prefix is disallowed.
    "note": a note that conveys anything important beyond description - what is noteworthy or weird about this unit or its usage. Implemented as an object of strings for all language mutations.
'''

units_dict_list = [
    # Eight basic units
    {"v": [1, 0, 0, 0, 0, 0, 0, 0], "id": "m", "name": {"cz": "metr", "en": "metre", "ae": "meter"}, "k": 1, "SI": True, "basic": True, "prefix": "all"},
    {"v": [0, 1, 0, 0, 0, 0, 0, 0], "id": "kg", "name": {"cz": "kilogram", "en": "kilogram"}, "k": 1, "SI": True, "basic": True, "note": {
        "cz": "To protože kilogram se obtížně programuje, neboť samo 'kilo' je předpona. Proto jsem definoval také gram jako odvozenou jednotku SI, která může mít jakékoliv předpony.",
        "en": "That's because kilogram is problematic to code, since the 'kilo' itself is a prefix. Therefore I have also defined gram as a derived SI unit, which can have all prefixes."
    }},
    {"v": [0, 0, 1, 0, 0, 0, 0, 0], "id": "s", "name": {"cz": "sekunda", "en": "second"}, "k": 1, "SI": True, "basic": True, "prefix": "-"},
    {"v": [0, 0, 0, 1, 0, 0, 0, 0], "id": "A", "name": {"cz": "ampér", "en": "ampere"}, "k": 1, "SI": True, "basic": True, "prefix": "all"},
    {"v": [0, 0, 0, 0, 1, 0, 0, 0], "id": "K", "name": {"cz": "kelvin", "en": "kelvin"}, "k": 1, "SI": True, "basic": True, "prefix": "all"},
    {"v": [0, 0, 0, 0, 0, 1, 0, 0], "id": "mol", "name": {"cz": "mol", "en": "mole"}, "k": 1, "SI": True, "basic": True, "prefix": "all"},
    {"v": [0, 0, 0, 0, 0, 0, 1, 0], "id": "cd", "name": {"cz": "kandela", "en": "candela"}, "k": 1, "SI": True, "basic": True, "prefix": "all"},
    # USD arbitrarily set as basic unit. Reference to this unit is hardcoded in currency loading!
    {"v": [0, 0, 0, 0, 0, 0, 0, 1], "id": "USD", "alias": ["$", "usd"], "name": {"cz": "americký dolar", "en": "US dollar"}, "k": 1, "basic": True, "prefix": "+"},


    #ALL OTHER UNITS as {id: 'identifier',v: [0,0,0,0,0,0,0], "name": {"cz": 'CZ', "en": 'EN'}, "k":1, "SI": True, "prefix": 'all'},
    #SI derived
    {"v": [0,0,0,0,0,0,0,0], "id": '%', "name": {"cz": 'procento', "en": 'percent'}, "k":1e-2},
    {"v": [0,0,0,0,0,0,0,0], "id": 'ppm', "name": {"cz": 'dílů na jeden milion', "en": 'parts per million'}, "k":1e-6},
    {"v": [0,0,0,0,0,0,0,0], "id": 'ppb', "name": {"cz": 'dílů na jednu miliardu', "en": 'parts per billion'}, "k":1e-9},
    {"v": [0,0,0,0,0,0,0,0], "id": 'rad', "name": {"cz": 'radián', "en": 'radian'}, "k":1, "SI": True, "prefix": '-', "note": {
        "cz": 'Úhel považuji za bezrozměrné číslo, čili radián je identický s číslem 1.',
        "en": 'I consider angle units to be dimensionless, with radian being identical to number 1.'}},
    {"v": [0,0,0,0,0,0,0,0], "id": '°', "alias":['deg'], "name": {"cz": 'stupeň', "en": 'degree'}, "k":Math.pi/180},
    {"v": [0,0,0,0,0,0,0,0], "id": 'gon', "name": {"cz": 'gradián', "en": 'gradian'}, "k":Math.pi/200},
    {"v": [0,0,-1,0,0,0,0,0], "id": 'Hz', "name": {"cz": 'hertz', "en": 'hertz'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [1,1,-2,0,0,0,0,0], "id": 'N', "name": {"cz": 'newton', "en": 'newton'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [-1,1,-2,0,0,0,0,0], "id": 'Pa', "name": {"cz": 'pascal', "en": 'pascal'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'J', "name": {"cz": 'joule', "en": 'joule'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,1,-3,0,0,0,0,0], "id": 'W', "name": {"cz": 'watt', "en": 'watt'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [0,0,1,1,0,0,0,0], "id": 'C', "name": {"cz": 'coulomb', "en": 'coulomb'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,1,-3,-1,0,0,0,0], "id": 'V', "name": {"cz": 'volt', "en": 'volt'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [-2,-1,4,2,0,0,0,0], "id": 'F', "name": {"cz": 'farad', "en": 'farad'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,1,-3,-2,0,0,0,0], "id": 'ohm', "name": {"cz": 'ohm', "en": 'ohm'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [-2,-1,3,2,0,0,0,0], "id": 'S', "name": {"cz": 'siemens', "en": 'siemens'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,1,-2,-1,0,0,0,0], "id": 'Wb', "name": {"cz": 'weber', "en": 'weber'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [0,1,-2,-1,0,0,0,0], "id": 'T', "name": {"cz": 'tesla', "en": 'tesla'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,1,-2,-2,0,0,0,0], "id": 'H', "name": {"cz": 'henry', "en": 'henry'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [0,0,0,0,1,0,0,0], "id": '°C', "name": {"cz": 'stupeň Celsia', "en": 'degree Celsius'}, "k":1, "SI": True, "note": {
        "cz": '°C je považován za jednotku rozdílu teploty (ΔT). Absolutní teplota (T) je zapsána pomocí složených závorek, např. {10°C}, viz tutoriál',
        "en": '°C is considered to be a unit of temperature difference (ΔT). Absolute temperature (T) is written using curly braces, e.g. {10°C}, see tutorial'}},

    {"v": [0,0,0,0,0,0,1,0], "id": 'lm', "name": {"cz": 'lumen', "en": 'lumen'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [-2,0,0,0,0,0,1,0], "id": 'lx', "name": {"cz": 'lux', "en": 'lux'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [0,0,-1,0,0,0,0,0], "id": 'Bq', "name": {"cz": 'becquerel', "en": 'becquerel'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [0,0,-1,0,0,0,0,0], "id": 'Rd', "name": {"cz": 'rutherford', "en": 'rutherford'}, "k":1e6, "SI": True, "prefix": 'all'},
    {"v": [2,0,-2,0,0,0,0,0], "id": 'Gy', "name": {"cz": 'gray', "en": 'gray'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [2,0,-2,0,0,0,0,0], "id": 'Sv', "name": {"cz": 'sievert', "en": 'sievert'}, "k":1, "SI": True, "prefix": 'all'},
    {"v": [-1,0,0,0,0,0,0,0], "id": 'dpt', "name": {"cz": 'dioptrie', "en": 'dioptre', "ae":'diopter'}, "k":1, "SI": True},



    #non-SI
    {"v": [0,0,0,0,0,1,0,0], "id": 'Nm3', "alias":['Ncm'], "name": {"cz": 'normální krychlový metr', "en": 'normal cubic metre'}, "k":csts.atm/csts.TC0/csts.R, "note": {
        "cz": 'Definován při 0°C a 1 atm. Navzdory názvu je Nm3 jednotkou látkového množství, nikoliv objemu.',
        "en": 'Defined at 0°C and 1 atm. Despite the name, Nm3 is actually amount of substance, not volume.'}},
    {"v": [0,0,0,0,0,1,0,0], "id": 'Ndm3', "alias":['Nl'], "name": {"cz": 'normální litr', "en": 'normal litre', 'ae':"normal liter"}, "k":csts.atm/csts.TC0/csts.R/1000, "note": {"cz": 'Viz Nm3 pro vysvětlení.', "en": 'See Nm3 for explanation.'}},
    {"v": [0,0,0,0,0,1,0,0], "id": 'SCF', "name": {"cz": 'normální krychlová stopa', "en": 'standard cubic foot'}, "k":0.028317*csts.atm/288.7/csts.R, "note": {"cz": 'Viz Nm3 pro vysvětlení.', "en": 'See Nm3 for explanation.'}},

    {"v": [0,0,1,0,0,0,0,0], "id": 'min', "name": {"cz": 'minuta', "en": 'minute'}, "k":60},
    {"v": [0,0,1,0,0,0,0,0], "id": 'h', "name": {"cz": 'hodina', "en": 'hour'}, "k":3600},
    {"v": [0,0,1,0,0,0,0,0], "id": 'd', "alias":['day'], "name": {"cz": 'den', "en": 'day'}, "k":3600*24},
    {"v": [0,0,1,0,0,0,0,0], "id": 'week', "name": {"cz": 'týden', "en": 'week'}, "k":3600*24*7},
    {"v": [0,0,1,0,0,0,0,0], "id": 'ftn', "name": {"cz": 'dva týdny', "en": 'fortnight'}, "k":1209600, "prefix": 'all'},
    {"v": [0,0,1,0,0,0,0,0], "id": 'month', "alias":['mth', 'měs'], "name": {"cz": 'průměrný měsíc', "en": 'average month'}, "k":3600*24*30.436875, "note": {
        "cz": 'Vypočten z gregoriánského roku.',
        "en": 'Calculated from gregorian year.'}},
    {"v": [0,0,1,0,0,0,0,0], "id": 'yr', "alias":['year'], "name": {"cz": 'gregoriánský rok', "en": 'gregorian year'}, "k":3600*24*365.2425, "note": {
        "cz": 'Pokud si nejste jisti, který rok použít, zvolte tento. Juliánský rok je zastaralý.',
        "en": 'If you are unsure which year to use, pick this one. Julian year is obsolete.'}},
    {"v": [0,0,1,0,0,0,0,0], "id": 'jyr', "name": {"cz": 'juliánský rok', "en": 'julian year'}, "k":3600*24*365.25},

    {"v": [0,0,-1,0,0,0,0,0], "id": 'rpm', "name": {"cz": 'otáčky za minutu', "en": 'revolutions per minute'}, "k":1/60},

    {"v": [0,0,0,0,1,0,0,0], "id": '°F', "name": {"cz": 'stupeň Fahrenheita', "en": 'degree Fahrenheit'}, "k":5/9, "note": csts.dTnote},
    {"v": [0,0,0,0,1,0,0,0], "id": '°Re', "alias":['°Ré', '°r'], "name": {"cz": 'stupeň Réaumura', "en": 'degree Réaumur'}, "k":1.25, "note": csts.dTnote},
    {"v": [0,0,0,0,1,0,0,0], "id": '°R', "name": {"cz": 'Rankine', "en": 'Rankine'}, "k":5/9, "note": csts.dTnote},

    {"v": [1,0,0,0,0,0,0,0], "id": 'Å', "name": {"cz": 'angstrom', "en": 'angstrom'}, "k":1e-10, "SI": True},
    {"v": [1,0,0,0,0,0,0,0], "id": 'th', "name": {"cz": 'thou', "en": 'thou'}, "k":2.54e-5},
    {"v": [1,0,0,0,0,0,0,0], "id": 'in', "name": {"cz": 'palec', "en": 'inch'}, "k":2.54e-2},
    {"v": [1,0,0,0,0,0,0,0], "id": 'ft', "name": {"cz": 'stopa', "en": 'foot'}, "k":0.3048, "prefix": '+'},
    {"v": [1,0,0,0,0,0,0,0], "id": 'yd', "name": {"cz": 'yard', "en": 'yard'}, "k":0.9144},
    {"v": [1,0,0,0,0,0,0,0], "id": 'fur', "name": {"cz": 'furlong', "en": 'furlong'}, "k":201.168, "prefix": 'all'},
    {"v": [1,0,0,0,0,0,0,0], "id": 'mi', "name": {"cz": 'míle', "en": 'mile'}, "k":csts.mile},
    {"v": [1,0,0,0,0,0,0,0], "id": 'nmi', "name": {"cz": 'námořní míle', "en": 'nautical mile'}, "k":1852},
    {"v": [1,0,0,0,0,0,0,0], "id": 'au', "name": {"cz": 'astronomická jednotka', "en": 'astronomical unit'}, "k":149597870700, "prefix": '+'},
    {"v": [1,0,0,0,0,0,0,0], "id": 'pc', "name": {"cz": 'parsek', "en": 'parsec'}, "k":3.08567758149137e16, "prefix": '+'},
    {"v": [1,0,0,0,0,0,0,0], "id": 'ly', "name": {"cz": 'světelný rok', "en": 'light-year'}, "k":9460730472580800, "prefix": '+'},

    {"v": [2,0,0,0,0,0,0,0], "id": 'a', "name": {"cz": 'ar', "en": 'ar'}, "k":100, "SI": True, "prefix": '+'},
    {"v": [2,0,0,0,0,0,0,0], "id": 'ac', "name": {"cz": 'akr', "en": 'acre'}, "k":4046.872},
    {"v": [2,0,0,0,0,0,0,0], "id": 'darcy', "name": {"cz": 'darcy', "en": 'darcy'}, "k": 9.869233e-13},

    {"v": [3,0,0,0,0,0,0,0], "id": 'l', "name": {"cz": 'litr', "en": 'litre', "ae":'liter'}, "k":1e-3, "SI": True, "prefix": 'all'},
    {"v": [3,0,0,0,0,0,0,0], "id": 'pt', "name": {"cz": 'pinta', "en": 'pint'}, "k":568.261e-6},
    {"v": [3,0,0,0,0,0,0,0], "id": 'gal', "name": {"cz": 'americký galon', "en": 'US gallon'}, "k":3.785412e-3},
    {"v": [3,0,0,0,0,0,0,0], "id": 'bsh', "name": {"cz": 'americký bušl', "en": 'US bushel'}, "k":35.2391e-3},
    {"v": [3,0,0,0,0,0,0,0], "id": 'ccm', "name": {"cz": 'kubický centimetr', "en": 'cubic centimetr', "ae": 'cubic centimeter'}, "k":1e-6},
    {"v": [3,0,0,0,0,0,0,0], "id": 'bbl', "name": {"cz": 'barel ropy', "en": 'oil barrel'}, "k":csts.bbl, "prefix": '+'},

    {"v": [3,0,-1,0,0,0,0,0], "id": 'BPD', "name": {"cz": 'barel ropy za den', "en": 'oil barrel per day'}, "k":csts.bbl/3600/24, "prefix": '+'},

    {"v": [0,1,0,0,0,0,0,0], "id": 'g', "name": {"cz": 'gram', "en": 'gram'}, "k":1e-3, "SI": True, "prefix": 'all'},
    {"v": [0,1,0,0,0,0,0,0], "id": 't', "name": {"cz": 'tuna', "en": 'tonne', "ae":'metric ton'}, "k":1000, "SI": True, "prefix": '+'},
    {"v": [0,1,0,0,0,0,0,0], "id": 'gr', "name": {"cz": 'grain', "en": 'grain'}, "k":64.79891e-6},
    {"v": [0,1,0,0,0,0,0,0], "id": 'oz', "name": {"cz": 'once', "en": 'ounce'}, "k":28.349523e-3},
    {"v": [0,1,0,0,0,0,0,0], "id": 'ozt', "name": {"cz": 'trojská unce', "en": 'troy ounce'}, "k":31.1034768e-3},
    {"v": [0,1,0,0,0,0,0,0], "id": 'ct', "name": {"cz": 'karát', "en": 'carat'}, "k":200e-6},
    {"v": [0,1,0,0,0,0,0,0], "id": 'lb', "alias":['lbs'], "name": {"cz": 'libra', "en": 'pound'}, "k":0.45359237},
    {"v": [0,1,0,0,0,0,0,0], "id": 'st', "name": {"cz": 'kámen', "en": 'stone'}, "k":6.35029318},
    {"v": [0,1,0,0,0,0,0,0], "id": 'slug', "name": {"cz": 'slug', "en": 'slug'}, "k":14.593903},
    {"v": [0,1,0,0,0,0,0,0], "id": 'fir', "name": {"cz": 'firkin', "en": 'firkin'}, "k":40.8233133, "prefix": 'all'},
    {"v": [0,1,0,0,0,0,0,0], "id": 'ts', "name": {"cz": 'krátká tuna', "en": 'short ton'}, "k":907.18474},
    {"v": [0,1,0,0,0,0,0,0], "id": 'tl', "name": {"cz": 'imperiální tuna', "en": 'long ton'}, "k":1016},
    {"v": [0,1,0,0,0,0,0,0], "id": 'u', "alias":['Da'], "name": {"cz": 'dalton (atomová hmotnostní konstanta)', "en": 'dalton (unified atomic mass unit)'}, "k":1.660539040e-27},

    {"v": [1,0,-1,0,0,0,0,0], "id": 'mph', "name": {"cz": 'míle za hodinu', "en": 'mile per hour'}, "k":csts.mile/3600},
    {"v": [1,0,-1,0,0,0,0,0], "id": 'kn', "name": {"cz": 'uzel', "en": 'knot'}, "k":1852/3600},

    {"v": [1,1,-2,0,0,0,0,0], "id": 'dyn', "name": {"cz": 'dyn', "en": 'dyne'}, "k":1e-5, "prefix": 'all'},

    {"v": [2,1,-2,0,0,0,0,0], "id": 'Wh', "name": {"cz": 'watthodina', "en": 'watt-hour'}, "k":3600, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'eV', "name": {"cz": 'elektronvolt', "en": 'electronvolt'}, "k":csts.q, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'erg', "name": {"cz": 'erg', "en": 'erg'}, "k":1e-7, "SI": True, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'Btu', "alias":['BTU','btu'], "name": {"cz": 'britská tepelná jednotka', "en": 'british thermal unit'}, "k":csts.BTU, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'Chu', "alias":['CHU','chu'], "name": {"cz": 'celsiova jednotka tepla', "en": 'celsius heat unit'}, "k": 1.899101e3, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'thm', "name": {"cz": 'therm', "en": 'therm'}, "k":csts.BTU*1e5, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'cal', "name": {"cz": 'kalorie', "en": 'calorie'}, "k":4.184, "prefix": 'all'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'TNT', "name": {"cz": 'ekvivalent tuny TNT', "en": 'ton of TNT equivalent'}, "k":4.184e9, "prefix": '+'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'BOE', "alias": ['BFOE'], "name": {"cz": 'ekvivalent barelu ropy', "en": 'barrel of oil equivalent'}, "k":5.8e6*csts.BTU, "prefix": '+'},
    {"v": [2,1,-2,0,0,0,0,0], "id": 'GGE', "name": {"cz": 'ekvivalent galonu benzínu', "en": 'gasoline gallon equivalent'}, "k":114e3*csts.BTU, "prefix": '+'},

    {"v": [-1,0,0,1,0,0,0,0], "id": 'Oe', "name": {"cz": 'oersted', "en": 'oersted'}, "k":1000/(4*Math.pi), "prefix": 'all'},

    {"v": [2,1,-3,0,0,0,0,0], "id": 'hp', "name": {"cz": 'imperiální koňská síla', "en": 'imperial horsepower'}, "k":745.69987158227022},

    {"v": [-1,1,-1,0,0,0,0,0], "id": 'P', "name": {"cz": 'poise', "en": 'poise'}, "k":0.1, "SI": True, "prefix": 'all'},
    {"v": [2,0,-1,0,0,0,0,0], "id": 'St', "name": {"cz": 'stokes', "en": 'stokes'}, "k":1e-4, "SI": True, "prefix": 'all'},

    {"v": [-1,1,-2,0,0,0,0,0], "id": 'bar', "name": {"cz": 'bar', "en": 'bar'}, "k":1e5, "SI": True, "prefix": 'all'},
    {"v": [-1,1,-2,0,0,0,0,0], "id": 'atm', "name": {"cz": 'atmosféra', "en": 'atmosphere'}, "k":csts.atm, "note": {
        "cz": 'Také slouží jako standardní tlak.',
        "en": 'Also serves as standard pressure.'}},
    {"v": [-1,1,-2,0,0,0,0,0], "id": 'mmHg', "name": {"cz": 'milimetr rtuťového sloupce', "en": 'millimetre of mercury', "ae": 'millimeter of mercury'}, "k":133.322387415, "note": csts.Hgnote},
    {"v": [-1,1,-2,0,0,0,0,0], "id": 'Torr', "alias":['torr'], "name": {"cz": 'torr', "en": 'torr'}, "k":csts.atm/760, "prefix": 'all', "note": csts.Hgnote},
    {"v": [-1,1,-2,0,0,0,0,0], "id": 'psi', "name": {"cz": 'libra na čtvereční palec', "en": 'pound per square inch'}, "k":6894.757293168362, "prefix": 'all'},

    {"v": [0,1,-2,-1,0,0,0,0], "id": 'G', "name": {"cz": 'gauss', "en": 'gauss'}, "k":0.0001, "SI": True, "prefix": 'all'},

    {"v": [0,0,-1,0,0,0,0,0], "id": 'Ci', "name": {"cz": 'Curie', "en": 'Curie'}, "k":3.7e10, "SI": False, "prefix": 'all'},
    {"v": [0,-1,1,1,0,0,0,0], "id": 'R', "name": {"cz": 'Rentgen', "en": 'Roentgen'}, "k":2.58e-4, "SI": False, "prefix": 'all'},
    {"v": [0,0,0,0,0,0,0,0], "id": 'monolayer',  "alias":['monolayers'], "name": {"cz": 'monolayer', "en": 'monolayer'}, "k":1},



    #constants
    {"v": [1,0,-2,0,0,0,0,0], "id": '_g', "name": {"cz": 'normální tíhové zrychlení', "en": 'standard gravity'}, "k":9.80665, "constant": True, "note": {
        "cz": 'Nikoliv univerzální konstanta, nýbrž konvenční.',
        "en": 'Not a universal constant, but a conventional one.'}},
    {"v": [1,0,-1,0,0,0,0,0], "id": '_c', "name": {"cz": 'rychlost světla ve vakuu', "en": 'speed of light in vacuum'}, "k":299792458, "constant": True},
    {"v": [3,-1,-2,0,0,0,0,0], "id": '_G', "name": {"cz": 'gravitační konstanta', "en": 'gravitational constant'}, "k":6.67408e-11, "constant": True},
    {"v": [2,1,-1,0,0,0,0,0], "id": '_h', "name": {"cz": 'Planckova konstanta', "en": 'Planck constant'}, "k":6.626070040e-34, "constant": True},
    {"v": [2,1,-2,0,-1,0,0,0], "id": '_k', "name": {"cz": 'Boltzmannova konstanta', "en": 'Boltzmann constant'}, "k":1.38064852e-23, "constant": True},
    {"v": [2,1,-2,0,-1,-1,0,0], "id": '_R', "name": {"cz": 'plynová konstanta', "en": 'gas constant'}, "k":csts.R, "constant": True},
    {"v": [1,1,-2,-2,0,0,0,0], "id": '_mu', "alias":['μ'], "name": {"cz": 'permeabilita vakua', "en": 'vacuum permeability'}, "k":1.2566370614e-6, "constant": True},
    {"v": [-3,-1,4,2,0,0,0,0], "id": '_E', "name": {"cz": 'permitivita vakua', "en": 'vacuum permittivity'}, "k":8.854187817e-12, "constant": True},
    {"v": [0,0,1,1,0,0,0,0], "id": '_q', "name": {"cz": 'elementární náboj', "en": 'elementary charge'}, "k":csts.q, "constant": True},
    {"v": [0,0,0,0,0,-1,0,0], "id": '_NA', "name": {"cz": 'Avogadrova konstanta', "en": 'Avogadro constant'}, "k":6.02214085e23, "constant": True},
    {"v": [0,0,0,0,0,0,0,0], "id": '_pi', "alias":['π'], "name": {"cz": 'Ludolfovo číslo', "en": 'Archimedes\' constant'}, "k":Math.pi, "constant": True},
    {"v": [0,0,0,0,0,0,0,0], "id": '_e', "name": {"cz": 'Eulerovo číslo', "en": 'Euler\'s number'}, "k":Math.e, "constant": True},
#     #special for Unitfuns, unusable without {}
#     {"v": [3,-1,0,0,0,0,0,0], "id": 'API', "alias":['°API'], "name": {"cz": 'API hustota', "en": 'API density'}, "k":1/141.5e3, onlyUnitfuns: True}, #theoretically, without {} API = specific volume unit
#     {"v": [0,0,0,0,0,0,0,0], "id": 'ln', "alias":['log'], "name": {"cz": 'Přirozený logaritmus', "en": 'Natural logarithm'}, "k":NaN, onlyUnitfuns: True}
];

# #unitfuns - irregular units that have a conversion function instead of mere ratio
# #{id: link to regular unit, f: function UF => SI, fi: inverse function SI => UF, "v": SI dimension (output when f, input when fi)}
# dict_Unitfuns = [
#     {id: '°C', f: UF => UF + csts.TC0, fi: SI => SI - csts.TC0, "v": [0,0,0,0,1,0,0,0]},
#     {id: '°F', f: UF => 5/9*UF + csts.TF0, fi: SI => 9/5*(SI - csts.TF0), "v": [0,0,0,0,1,0,0,0]},
#     {id: '°Re', f: UF => 1.25*UF + csts.TC0, fi: SI => 0.8*(SI - csts.TC0), "v": [0,0,0,0,1,0,0,0]},
#     {id: 'API', f: UF => csts.APIk/(UF - csts.APIq), fi: SI => csts.APIk/SI + csts.APIq, "v": [-3,1,0,0,0,0,0,0]},
#     {id: 'ln', f: UF => Math.log(UF), fi: SI => {throw '🏆 '+langService.trans('ERR_Secret');}, "v": [0,0,0,0,0,0,0,0]}
# ];

#currencies - their conversion ratio to dollar is unknown and will be obtained by currencies.php
#k and v will be filled later (v is always the same, k is obtained from API)
dict_Currencies = [
    {id: 'EUR', "alias":['€'], "name": {"cz": 'euro', "en": 'Euro'}},
    {id: 'AED', "name": {"cz": 'dirham Spojených arabských emirátů', "en": 'United Arab Emirates Dirham'}},
    {id: 'ARS', "name": {"cz": 'argentinské peso', "en": 'Argentine Peso'}},
    {id: 'AUD', "name": {"cz": 'australský dolar', "en": 'Australian Dollar'}},
    {id: 'BGN', "name": {"cz": 'bulharský lev', "en": 'Bulgarian Lev'}},
    {id: 'BRL', "name": {"cz": 'braziliský real', "en": 'Brazilian Real'}},
    {id: 'CAD', "name": {"cz": 'kanadský dolar', "en": 'Canadian Dollar'}},
    {id: 'CHF', "name": {"cz": 'švýcarský frank', "en": 'Swiss Franc'}},
    {id: 'CNY', "name": {"cz": 'čínský juan', "en": 'Chinese Yuan'}},
    {id: 'CZK', "alias":['Kč'], "name": {"cz": 'česká koruna', "en": 'Czech Republic Koruna'}},
    {id: 'DKK', "name": {"cz": 'dánská koruna', "en": 'Danish Krone'}},
    {id: 'GBP', "alias":['£'], "name": {"cz": 'britská libra', "en": 'British Pound Sterling'}},
    {id: 'HKD', "name": {"cz": 'hongkongský dolar', "en": 'Hong Kong Dollar'}},
    {id: 'HRK', "name": {"cz": 'chorvatská kuna', "en": 'Croatian Kuna'}},
    {id: 'HUF', "name": {"cz": 'maďarský forint', "en": 'Hungarian Forint'}},
    {id: 'IDR', "name": {"cz": 'indonéská rupie', "en": 'Indonesian Rupiah'}},
    {id: 'ILS', "name": {"cz": 'nový izraelský šekel', "en": 'Israeli New Sheqel'}},
    {id: 'INR', "name": {"cz": 'indická rupie', "en": 'Indian Rupee'}},
    {id: 'JPY', "alias": ['¥'], "name": {"cz": 'japonský jen', "en": 'Japanese Yen'}},
    {id: 'KRW', "name": {"cz": 'jihokorejský won', "en": 'South Korean Won'}},
    {id: 'MXN', "name": {"cz": 'mexické peso', "en": 'Mexican Peso'}},
    {id: 'NOK', "name": {"cz": 'norská koruna', "en": 'Norwegian Krone'}},
    {id: 'NZD', "name": {"cz": 'novozélandský dolar', "en": 'New Zealand Dollar'}},
    {id: 'PLN', "name": {"cz": 'polský zlotý', "en": 'Polish Zloty'}},
    {id: 'RON', "name": {"cz": 'rumunské leu', "en": 'Romanian Leu'}},
    {id: 'RUB', "name": {"cz": 'ruský rubl', "en": 'Russian Ruble'}},
    {id: 'SEK', "name": {"cz": 'švédská koruna', "en": 'Swedish Krona'}},
    {id: 'SGD', "name": {"cz": 'singapurský dolar', "en": 'Singapore Dollar'}},
    {id: 'THB', "name": {"cz": 'thajský baht', "en": 'Thai Baht'}},
    {id: 'TRY', "name": {"cz": 'turecká lira', "en": 'Turkish Lira'}},
    {id: 'VND', "name": {"cz": 'vietnamský dong', "en": 'Vietnamese Dong'}},
    {id: 'ZAR', "name": {"cz": 'jihoafrický rand', "en": 'South African Rand'}},
    {id: 'BTC', "name": {"cz": 'bitcoin', "en": 'Bitcoin'}}
];

#standard SI prefixes
dict_Prefixes = [
    {id: 'a', "v": -18},
    {id: 'f', "v": -15},
    {id: 'p', "v": -12},
    {id: 'n', "v": -9},
    {id: 'u', "v": -6},
    {id: 'μ', "v": -6},
    {id: 'm', "v": -3},
    {id: 'c', "v": -2},
    {id: 'd', "v": -1},
    {id: 'h', "v": 2},
    {id: 'k', "v": 3},
    {id: 'M', "v": 6},
    {id: 'G', "v": 9},
    {id: 'T', "v": 12},
    {id: 'P', "v": 15}
];

# Define the list of dictionaries
#Extract names in the specified order
def extract_names_by_order(units_dict_list, order):
    ordered_names = []
    for unit in units_dict_list:
        if "name" in unit:
            for key in order:
                if key in unit["name"]:
                    ordered_names.append(unit["name"][key])
    return ordered_names

def make_ids_dict(units_dict_list):
    ids_dict = {}
    for unit_dict in units_dict_list:
        if "id" in unit_dict:
            ids_dict[unit_dict["id"]] = unit_dict
    return ids_dict

# Specify the order of keys
name_order = ["ae", "en", "cz"]

#metric prefixes:
metric_prefixes = ["yotta", "zetta", "exa", "peta", "tera", "giga", "mega", "kilo", "hecto", "deca", "", "deci", "centi", "milli", "micro", "nano", "pico", "femto", "atto", "zepto", "yocto"]
metric_prefixes_greater = metric_prefixes[0:11] #first 10 
metric_prefixes_lesser = metric_prefixes[11:]
metric_prefixes_symbolic = ["Y", "Z", "E", "P", "T", "G", "M", "k", "h", "da", "", "d", "c", "m", "µ", "n", "p", "f", "a", "z", "y"]
metric_prefixes_symbolic_greater = metric_prefixes_symbolic[0:11] #first 10 
metric_prefixes_symbolic_lesser = metric_prefixes_symbolic[11:]

# Extract and print the names
ordered_name_list = extract_names_by_order(units_dict_list, name_order)
ids_dict = make_ids_dict(units_dict_list)

#now expand them with the metric prefix permutations
def expand_ids_list(ids_list, ids_dict):
    """
    Gives back the expanded names list by generating permutations with metric prefixes.
    Such as:
    ["km", "cm", "mm"]
    """
    expanded_ids_list = [] #now create the list to add to.
    for id in ids_list:
        if "prefix" in ids_dict[id]:
            if ids_dict[id]["prefix"] == "all":
                for metric_prefix_symbolic in metric_prefixes_symbolic:
                    expanded_ids_list.append(metric_prefix_symbolic+id)
            elif ids_dict[id]["prefix"] == "-":
                expanded_ids_list.append(id) #no prefix
                for metric_prefix_symbolic in metric_prefixes_symbolic_lesser:
                    expanded_ids_list.append(metric_prefix_symbolic+id)
            elif ids_dict[id]["prefix"] == "+":
                expanded_ids_list.append(id) #no prefix
                for metric_prefix_symbolic in metric_prefixes_symbolic_greater:
                    expanded_ids_list.append(metric_prefix_symbolic+id)
            else:
                expanded_ids_list.append(id) #no prefix
        else:
                expanded_ids_list.append(id) #no prefix
    return expanded_ids_list
ids_list = list(ids_dict.keys())
expanded_ids_list = expand_ids_list(ids_list, ids_dict)
expanded_ids_set = set(expanded_ids_list)

def expand_names_list(ids_list, ids_dict):
    """
    Gives back the expanded names list by generating permutations with metric prefixes.
    Such as:
    ["kilometer", "megameter", "gigameter"]
    
    The lines which say ids_dict[id]["alias"][0] could loop across aliases.
    However, there is usually only one alias, and the first is usually sufficient when there is more than one.
    """
    expanded_names_list = []  # Create the list to add to.
    for id in ids_list:
        if "prefix" in ids_dict[id]:
            if ids_dict[id]["prefix"] == "all": #for "all" prefix.
                #loop across prefixes, includes base case.
                for metric_prefix in metric_prefixes:  # Use "metric_prefixes".
                    expanded_names_list.append(metric_prefix + ids_dict[id]["name"]["en"])  # Use ids_dict[id]["name"]["en"] for concatenation.
                    if 'ae' in ids_dict[id]["name"]:
                        expanded_names_list.append(metric_prefix + ids_dict[id]["name"]["ae"]) 
                    if 'alias' in ids_dict[id]:
                        expanded_names_list.append(metric_prefix + ids_dict[id]["alias"][0])  
            elif ids_dict[id]["prefix"] == "-": #for "-" prefix.
                #base case
                expanded_names_list.append(ids_dict[id]["name"]["en"])  # No prefix.
                if 'ae' in ids_dict[id]["name"]:
                    expanded_names_list.append(ids_dict[id]["name"]["ae"])  # No prefix.
                if 'alias' in ids_dict[id]:
                    expanded_names_list.append(ids_dict[id]["alias"][0])  #No prefix.
                #loop across prefix cases
                for metric_prefix in metric_prefixes_lesser:  # Use lesser prefixes.
                    expanded_names_list.append(metric_prefix + ids_dict[id]["name"]["en"])
                    if 'ae' in ids_dict[id]["name"]:
                        expanded_names_list.append(metric_prefix + ids_dict[id]["name"]["ae"])
                    if 'alias' in ids_dict[id]:
                        expanded_names_list.append(metric_prefix + ids_dict[id]["alias"][0])
            elif ids_dict[id]["prefix"] == "+": #for "+" prefix.
                #base case
                expanded_names_list.append(ids_dict[id]["name"]["en"])  # No prefix.
                if 'ae' in ids_dict[id]["name"]:
                    expanded_names_list.append(ids_dict[id]["name"]["ae"])  # No prefix.
                if 'alias' in ids_dict[id]:
                    expanded_names_list.append(ids_dict[id]["alias"][0])
                #loop across prefixes
                for metric_prefix in metric_prefixes_greater:  # Use greater prefixes.
                    expanded_names_list.append(metric_prefix + ids_dict[id]["name"]["en"])
                    if 'ae' in ids_dict[id]["name"]:
                        expanded_names_list.append(metric_prefix + ids_dict[id]["name"]["ae"])
                    if 'alias' in ids_dict[id]:
                        expanded_names_list.append(metric_prefix + ids_dict[id]["alias"][0])
            else: #for odd prefix rules, we just use base case.
                #base case
                expanded_names_list.append(ids_dict[id]["name"]["en"])  # No prefix.
                if 'ae' in ids_dict[id]["name"]:
                    expanded_names_list.append(ids_dict[id]["name"]["ae"])  # No prefix.
                if 'alias' in ids_dict[id]:
                    expanded_names_list.append(ids_dict[id]["alias"][0])
        else: #for no prefix.
            expanded_names_list.append(ids_dict[id]["name"]["en"])  # No prefix.
            if 'ae' in ids_dict[id]["name"]:
                expanded_names_list.append(ids_dict[id]["name"]["ae"])  # No prefix.
            if 'alias' in ids_dict[id]:
                expanded_names_list.append(ids_dict[id]["alias"][0])
    return expanded_names_list


expanded_names_list = expand_names_list(ids_list, ids_dict)
expanded_names_set = set(expanded_names_list)


if __name__ == "__main__":
    #here is just the ordered name list.
    name_order = ["ae", "en", "cz"]
    ordered_name_list = extract_names_by_order(units_dict_list, name_order)
    print(ordered_name_list)
    #here is just the ordered name list for cz names.
    name_order = ["cz"]
    ordered_name_list = extract_names_by_order(units_dict_list, name_order)
    print(ordered_name_list)

    #here is just the ordered name list for en names.
    name_order = ["en"]
    ordered_name_list = extract_names_by_order(units_dict_list, name_order)
    print("\n", ordered_name_list)
    
    #print the ids_list:
    ids_list = make_ids_dict(units_dict_list)
    print(ids_list)



from sympy import symbols, Eq, solve, sympify
from pint import UnitRegistry
import re

ureg = UnitRegistry()

def parse_equation(equation_str, variables):
    """Replace variable names and standalone numbers with their magnitudes and formatted units."""
    def detect_and_format_units(equation_str):
        """Detect standalone numbers with units and format them correctly."""
        pattern = r"(\d+(\.\d+)?)\s*([a-zA-Z]+)"
        matches = re.findall(pattern, equation_str)
        # Explanation of regular expression parts
        # (\d+(\.\d+)?)
        #     \d+ → Matches one or more digits (e.g., "10", "100", "3").
        #     (\.\d+)? → Matches optional decimal values (e.g., "10.5", "3.14").
        #     This entire part captures numerical values, whether integers or decimals.
        # \s*
        #     Matches zero or more spaces between the number and the unit.
        #     Ensures flexibility in formatting (e.g., "10m" vs. "10 m").
        # ([a-zA-Z]+)
        #     Matches one or more alphabetical characters, capturing the unit symbol.
        #     Ensures that only valid unit names (e.g., "m", "s", "kg") are recognized.
        # Example Matches
        #     "10 m"   → ("10", "", "m")
        #     "3.5 kg" → ("3.5", ".5", "kg")
        #     "100s"   → ("100", "", "s")
        for match in matches:
            magnitude = match[0]  # Extract number
            unit = match[2]  # Extract unit
            try:
                quantity = ureg(f"{magnitude} {unit}")  # Convert to Pint quantity
                formatted_unit = str(quantity.units)
                equation_str = equation_str.replace(f"{magnitude} {unit}", f"({quantity.magnitude} * {formatted_unit})")
            except: #This comment is so that VS code pylint will not flag this line: pylint: disable=bare-except
                pass  # Ignore invalid unit conversions
        return equation_str


    # Sort variable names by length in descending order
    variables_sorted_by_name = sorted(variables.items(), key=lambda x: -len(x[0]))

    # Need to first replace the constants because they could be like "letter e"
    # and could mess up the string after the units are added in.
    # Also need to sort them by length and replace longer ones first because
    # we could have "Ea" and "a", for example.  
    for var_name, var_value in variables_sorted_by_name:  # Removed `.items()`
        if not hasattr(var_value, "magnitude"):  # For constants like "e" with no units  
            equation_str = equation_str.replace(var_name, str(var_value))  # Directly use plain values  

    # Replace variables with their magnitudes and units  
    for var_name, var_value in variables_sorted_by_name:  # Removed `.items()`
        if hasattr(var_value, "magnitude"):  # Ensure it has a magnitude attribute  
            magnitude = var_value.magnitude  
            unit = str(var_value.units)  
            equation_str = equation_str.replace(var_name, f"({magnitude} * {unit})")  

    # Detect and format standalone numbers with units, like "10 m"
    equation_str = detect_and_format_units(equation_str)
    return equation_str

def solve_equation(equation_string, independent_variables_values_and_units, dependent_variable):
    """
        Solve for the specified dependent variable in terms of multiple independent variables.
        # # Example usage
        # independent_variables_values_and_units = {
        #     "x": "2 m / s",
        #     "y": "3 meter"
        # }
        # equation_string = "x * t + y = 10 m"
        # solve_equation(equation_string, independent_variables_values_and_units, dependent_variable="t")
        # It will solve for the value of t. 
        # What is returned is a list of solutions.
        # if there is any "^" in the equation, it will be changed to **

    """
    # Convert string inputs into Pint quantities
    variables = {name: ureg(value) for name, value in independent_variables_values_and_units.items()}
    independent_variables = list(independent_variables_values_and_units.keys())
    # Explicitly define symbolic variables
    symbols_dict = {var: symbols(var) for var in independent_variables_values_and_units.keys()}
    for var in independent_variables:
        symbols_dict[var] = symbols(var)
    symbols_dict[dependent_variable] = symbols(dependent_variable)

    #change any "^" into "**"
    equation_string = equation_string.replace("^","**")
    # Split the equation into left-hand and right-hand sides
    lhs, rhs = equation_string.split("=")

    # Convert both sides to SymPy expressions
    lhs_sympy = sympify(parse_equation(lhs.strip(), variables), locals=symbols_dict, evaluate=False)
    rhs_sympy = sympify(parse_equation(rhs.strip(), variables), locals=symbols_dict, evaluate=False)

    # Create the equation object
    eq_sympy = Eq(lhs_sympy, rhs_sympy)
    # Solve for the dependent variable
    solutions = solve(eq_sympy, symbols_dict[dependent_variable])
    # Extract magnitude and unit separately from SymPy expressions
    separated_solutions = []
    for sol in solutions:
        magnitude, unit = sol.as_coeff_Mul()  # Works for ANY SymPy expression
        separated_solutions.append((magnitude, unit))

    # Format solutions properly with a space between the magnitude and unit
    formatted_solutions = [f"{mag} ({unit})" for mag, unit in separated_solutions]
    #print(f"Solutions for {dependent_variable} in terms of {independent_variables}: {formatted_solutions}")
    return formatted_solutions


def parse_equation_dict(equation_dict):
    def extract_value_units(entry):
        trimmed_entry = entry.strip()  # Remove leading/trailing whitespace
        split_entry = trimmed_entry.split(" ", 1)  # Split on the first space
        if len(split_entry) > 1:
            value = split_entry[0]
            units = split_entry[1] # Everything after the number
            return [value, units]
        else:
            return [float(split_entry[0]), None]  # Handle constants without units

    def extract_constants(constants_dict):
        return {
            name: extract_value_units(value)
            for name, value in constants_dict.items()
        }

    def extract_equation(equation_string):
        variables_list = re.findall(r"([A-Za-z]+)", equation_string)
        return {"equation_string": equation_string, "variables_list": variables_list}

    if 'graphical_dimensionality' in equation_dict:
        graphical_dimensionality = equation_dict['graphical_dimensionality']
    else:
        graphical_dimensionality = 2

    constants_extracted_dict = extract_constants(equation_dict["constants"])
    equation_extracted_dict = extract_equation(equation_dict["equation_string"])
    # x_match = re.match(r"([\w\d{}$/*_°α-ωΑ-Ω]+)\s*\(([\w\d{}$/*_°α-ωΑ-Ω]*)\)", equation_dict["x_variable"])
    # y_match = re.match(r"([\w\d{}$/*_°α-ωΑ-Ω]+)\s*\(([\w\d{}$/*_°α-ωΑ-Ω]*)\)", equation_dict["y_variable"])
    # x_match  = (x_match.group(1), x_match.group(2)) if x_match else (equation_dict["x_variable"], None)
    # y_match = (y_match.group(1), y_match.group(2)) if y_match else (equation_dict["y_variable"], None)
    x_match = extract_value_units(equation_dict["x_variable"])
    y_match = extract_value_units(equation_dict["y_variable"])
    if graphical_dimensionality == 3:
        z_match = extract_value_units(equation_dict["z_variable"])

    # Create dictionaries for extracted variables
    x_variable_extracted_dict = {"label": x_match[0], "units": x_match[1]}
    y_variable_extracted_dict = {"label": y_match[0], "units": y_match[1]}
    if graphical_dimensionality == 3:
        z_variable_extracted_dict = {"label": z_match[0], "units": z_match[1]}

    def prepare_independent_variables(constants_extracted_dict):
        independent_variables_dict = {
            name: f"{value} {units}" if units else f"{value}"
            for name, (value, units) in constants_extracted_dict.items()
        }
        return independent_variables_dict
    independent_variables_dict = prepare_independent_variables(constants_extracted_dict)

    if graphical_dimensionality == 2:
        return independent_variables_dict, constants_extracted_dict, equation_extracted_dict, x_variable_extracted_dict, y_variable_extracted_dict
    if graphical_dimensionality == 3:
        return independent_variables_dict, constants_extracted_dict, equation_extracted_dict, x_variable_extracted_dict, y_variable_extracted_dict, z_variable_extracted_dict

# equation_dict = {
#     'equation_string': 'k = A*(e**((-Ea)/(R*T)))',
#     'x_variable': 'T (K)',  
#     'y_variable': 'k (s**(-1))',
#     'constants': {'Ea': '30000 (J)*(mol^(-1))', 'R': '8.314 (J)*(mol^(-1))*(K^(-1))' , 'A': '1E13 (s**-1)', 'e': '2.71828'},
#     'num_of_points': 10,
#     'x_range_default': [200, 500],
#     'x_range_limits' : [],
#     'points_spacing': 'Linear'
# }

# try:
#     result_extracted = parse_equation_dict(equation_dict)
#     print(result_extracted)
# except ValueError as e:
#     print(f"Error: {e}")


def generate_multiplicative_points(range_min, range_max, num_of_points=None, factor=None, reverse_scaling=False):
    """
    Generates a sequence of points using relative spacing within a normalized range.
    
    - Spacing between points changes multiplicatively (e.g., doubling means each interval doubles).
    - Returns range_min and range_max explicitly in all cases.
    - Works for negative values and cases where min is negative while max is positive.
    - If `reverse_scaling` is True, exponential scaling occurs from the max end instead.

    Parameters:
    - range_min (float): The starting value of the sequence.
    - range_max (float): The maximum limit for generated values.
    - num_of_points (int, optional): Desired number of points (excluding min/max).
    - factor (float, optional): Multiplication factor for spacing between successive values.
    - reverse_scaling (bool, optional): If True, spacing is applied in reverse direction.

    Returns:
    - List of generated points (standard Python floats).

    Raises:
    - ValueError: If neither num_of_points nor factor is provided.
    """

    # Define normalized bounds
    relative_min = 0
    relative_max = 1
    total_value_range = range_max - range_min  

    # Case 1: num_of_points is provided (factor may be provided too)
    if num_of_points is not None and num_of_points > 1:
        
        # Case 1a: Generate points using equal spacing in relative space
        equal_spacing_list = [relative_min]  # Start at normalized min
        equal_spacing_value = (relative_max - relative_min) / (num_of_points - 1)  # Normalized step size

        for step_index in range(1, num_of_points):
            equal_spacing_list.append(relative_min + step_index * equal_spacing_value)

        # Case 1b: Generate points using multiplication factor (if provided)
        factor_spacing_list = [relative_min]
        if factor is not None and factor > 0:
            relative_spacing = 0.01  # Start at 1% of the range (normalized units)
            current_position = relative_min

            while current_position + relative_spacing < relative_max:
                current_position += relative_spacing
                factor_spacing_list.append(current_position)
                relative_spacing *= factor  # Multiply spacing by factor

        # Compare list lengths explicitly and select the better approach
        if len(factor_spacing_list) > len(equal_spacing_list):
            normalized_points = factor_spacing_list
        else:
            normalized_points = equal_spacing_list

    # Case 2: Only factor is provided, generate points using the multiplication factor
    elif factor is not None and factor > 0:
        relative_spacing = 0.01  # Start at 1% of the range
        current_position = relative_min
        normalized_points = [relative_min]

        while current_position + relative_spacing < relative_max:
            current_position += relative_spacing
            normalized_points.append(current_position)
            relative_spacing *= factor  # Multiply spacing by factor

    # Case 3: Neither num_of_points nor factor is provided, compute equal spacing dynamically
    elif num_of_points is None and factor is None:
        equal_spacing_value = (relative_max - relative_min) / 9  # Default to 9 intermediate points
        normalized_points = [relative_min + step_index * equal_spacing_value for step_index in range(1, 9)]

    # Case 4: Invalid input case—neither num_of_points nor factor is properly set
    else:
        raise ValueError("Either num_of_points or factor must be provided.")

    # Ensure the last relative point is relative_max before scaling
    if normalized_points[-1] != relative_max:
        normalized_points.append(relative_max)

    # Scale normalized points back to the actual range
    if reverse_scaling:
        scaled_points = [range_max - ((relative_max - p) * total_value_range) for p in normalized_points]  # Reverse scaling adjustment
    else:
        scaled_points = [range_min + (p * total_value_range) for p in normalized_points]

    return scaled_points

# # Example usages
# print("line 224")
# print(generate_multiplicative_points(0, 100, num_of_points=10, factor=2))  # Default exponential scaling from min end
# print(generate_multiplicative_points(0, 100, num_of_points=10, factor=2, reverse_scaling=True))  # Exponential scaling from max end
# print(generate_multiplicative_points(1, 100, num_of_points=10, factor=1.3)) # Compares num_of_points vs factor, chooses whichever makes more points
# print(generate_multiplicative_points(1, 100, num_of_points=10))            # Computes factor dynamically
# print(generate_multiplicative_points(1, 100, factor=2))                    # Uses factor normally
# print(generate_multiplicative_points(1, 100))                               # Uses step_factor for default 10 points with 8 intermediate values
# print("line 228")


# # Example usages with reverse scaling
# print("line 240")
# print(generate_multiplicative_points(-50, 100, num_of_points=10, factor=2))  # Case 1b: Uses spacing factor with num_of_points
# print(generate_multiplicative_points(-50, 100, num_of_points=10, factor=2, reverse_scaling=True))  # Reverse scaling version
# print(generate_multiplicative_points(-100, -10, num_of_points=5, factor=1.5)) # Case 1b: Works with negatives
# print(generate_multiplicative_points(-25, 75, num_of_points=7))               # Case 1a: Computes spacing dynamically
# print(generate_multiplicative_points(-10, 50, factor=1.3))                    # Case 2: Uses factor-based spacing
# print(generate_multiplicative_points(-30, 30))                                # Case 3: Uses default intermediate spacing

def generate_points_by_spacing(num_of_points=10, range_min=0, range_max=1, points_spacing="linear"):
    """
    Generates a sequence of points based on the specified spacing method.
    
    Supported spacing types:
    - "linear": Evenly spaced values between range_min and range_max.
    - "logarithmic": Logarithmically spaced values.
    - "exponential": Exponentially increasing values.
    - A real number > 0: Used as a multiplication factor to generate values.
    
    Parameters:
    - num_of_points (int): The number of points to generate. Default is 10.
    - range_min (float): The starting value of the sequence. Default is 1.
    - range_max (float): The maximum limit for generated values. Default is 100.
    - points_spacing (str or float): Defines the spacing method or multiplication factor.
    
    Returns:
    - List of generated points.
    
    Raises:
    - ValueError: If an unsupported spacing type is provided.
    """
    import numpy as np  # Ensure numpy is imported
    spacing_type = str(points_spacing).lower() if isinstance(points_spacing, str) else None
    points_list = None
    if num_of_points == None:
        num_of_points = 10
    if range_min == None:
        range_min = 0
    if range_max == None:
        range_max = 1
    if str(spacing_type).lower() == "none":
        spacing_type = "linear"
    if spacing_type == "":
        spacing_type = "linear"
    if spacing_type.lower() == "linear":
        points_list = np.linspace(range_min, range_max, num_of_points).tolist()
    elif spacing_type.lower() == "logarithmic":
        points_list = np.logspace(np.log10(range_min), np.log10(range_max), num_of_points).tolist()
    elif spacing_type.lower() == "exponential":
        points_list = (range_min * np.exp(np.linspace(0, np.log(range_max/range_min), num_of_points))).tolist()
    elif isinstance(points_spacing, (int, float)) and points_spacing > 0:
        points_list = generate_multiplicative_points(range_min, range_max, points_spacing, num_of_points)
    else:
        raise ValueError(f"Unsupported spacing type: {points_spacing}")

    return points_list


# # Example usage demonstrating different spacing types:
# print(generate_points_by_spacing(num_of_points=10, range_min=1, range_max=100, points_spacing="linear"))         # Linear spacing
# print(generate_points_by_spacing(num_of_points=10, range_min=1, range_max=100, points_spacing="logarithmic"))    # Logarithmic spacing
# print(generate_points_by_spacing(num_of_points=10, range_min=1, range_max=100, points_spacing="exponential"))    # Exponential spacing
# print(generate_points_by_spacing(num_of_points=10, range_min=1, range_max=100, points_spacing=2))               # Multiplicative factor spacing


def generate_points_from_range_dict(range_dict, variable_name="x"):
    """
    Extracts the necessary range and parameters from range_dict and generates a sequence of points.
    In practice, the range_dict can be a full equation_dict with extra fields that will not be used.

    The function follows these rules:
    1. If '{variable_name}_range_limits' is provided as a list of two numbers, it is used as the range.
    2. Otherwise, '{variable_name}_range_default' is used as the range.
    3. Calls generate_points_by_spacing() to generate the appropriate sequence based on num_of_points and points_spacing.

    Parameters:
    - range_dict (dict): Dictionary containing equation details, including range limits, num_of_points, and spacing type.
    - variable_name (str, optional): Name of the variable to determine the range settings. Defaults to 'x'.

    Returns:
    - List of generated points.
    """
    range_default_key = f"{variable_name}_range_default"
    range_limits_key = f"{variable_name}_range_limits"

    # Assigning range. 
    # Start with default values
    if range_dict.get(range_default_key):  # get prevents crashing if the field is not present.
        range_min, range_max = range_dict[range_default_key]

    # If '{variable_name}_range_limits' is provided, update values only if they narrow the range
    if range_dict.get(range_limits_key):
        limit_min, limit_max = range_dict[range_limits_key]
        # Apply limits only if they tighten the range
        if limit_min is not None and limit_min > range_min:
            range_min = limit_min
        if limit_max is not None and limit_max < range_max:
            range_max = limit_max

    # Ensure at least one valid limit exists
    if range_min is None or range_max is None:
        raise ValueError(f"At least one min and one max must be specified between {variable_name}_range_default and {variable_name}_range_limits.")

    list_of_points = generate_points_by_spacing(
        num_of_points=range_dict['num_of_points'],
        range_min=range_min,
        range_max=range_max,
        points_spacing=range_dict['points_spacing']
    )
    # Generate points using the specified spacing method
    return list_of_points


## Start of Portion of code for parsing out tagged ustom units and returning them ##

def return_custom_units_markup(units_string, custom_units_list):
    """puts markup around custom units with '<' and '>' """
    sorted_custom_units_list = sorted(custom_units_list, key=len, reverse=True)
    #the units should be sorted from longest to shortest if not already sorted that way.
    for custom_unit in sorted_custom_units_list:
        units_string = units_string.replace(custom_unit, '<'+custom_unit+'>')
    return units_string

def extract_tagged_strings(text):
    """Extracts tags surrounded by <> from a given string. Used for custom units.
       returns them as a list sorted from longest to shortest"""
    list_of_tags = re.findall(r'<(.*?)>', text)
    set_of_tags = set(list_of_tags)
    sorted_tags = sorted(set_of_tags, key=len, reverse=True)
    return sorted_tags

##End of Portion of code for parsing out tagged ustom units and returning them ##



#This function is to convert things like (1/bar) to (bar)**(-1)
#It was written by copilot and refined by further prompting of copilot by testing.
#The depth is because the function works iteratively and then stops when finished.
def convert_inverse_units(expression, depth=100):
    # Patterns to match valid reciprocals while ignoring multiplied units, so (1/bar)*bar should be  handled correctly.
    patterns = [r"1/\((1/.*?)\)", r"1/([a-zA-Z]+)"]
    for _ in range(depth):
        new_expression = expression
        for pattern in patterns:
            new_expression = re.sub(pattern, r"(\1)**(-1)", new_expression)
        
        # Stop early if no more changes are made
        if new_expression == expression:
            break
        expression = new_expression
    return expression

#This support function is just for code readability.
#It returnts two strings in a list, split at the first delimiter.
def split_at_first_delimiter(string, delimter=" "):
    return string.split(delimter, 1)

#This function takes an equation dict (see examples) and returns the x_points, y_points, and x_units and y_units.
#If there is more than one solution (like in a circle, for example) all solutions should be returned.
#The function is slow. I have checked what happens if "vectorize" is used on the x_point loop (which is the main work)
#and the function time didn't change. So the functions it calls must be where the slow portion is.
#I have not timed the individual functions to find and diagnose the slow step(s) to make them more efficient.
#Although there is lots of conversion between different object types to support the units format flexiblity that this function has,
#I would still expect the optimzed code to be an order of magnitude faster. So it may be worth finding the slow steps.
#One possibility might be to use "re.compile()"
def evaluate_equation_dict(equation_dict, verbose=False):
    import copy
    equation_dict = copy.deepcopy(equation_dict)  # Create a deep copy to prevent unintended modifications
    #First a block of code to extract the x_points needed
    # Extract each dictionary key as a local variable
    equation_string = equation_dict['equation_string']
    if 'graphical_dimensionality' in equation_dict:
        graphical_dimensionality = equation_dict['graphical_dimensionality']
        graphical_dimensionality_added = False
    else: #assume graphical_dimensionality is 2 if one is not provided.
        equation_dict['graphical_dimensionality'] = 2
        graphical_dimensionality_added = True
        graphical_dimensionality = 2
    if 'verbose' in equation_dict:
        verbose = equation_dict["verbose"]
    # We don't need the below variables, because they are in the equation_dict.
    # x_variable = equation_dict['x_variable']
    # y_variable = equation_dict['y_variable']
    # constants = equation_dict['constants']
    # reverse_scaling = equation_dict['reverse_scaling']
    x_points = generate_points_from_range_dict(range_dict = equation_dict, variable_name='x')
    if graphical_dimensionality == 3: #for graphical_dimensionality of 3, the y_points are also an independent_variable to generate.
        y_points = generate_points_from_range_dict(range_dict = equation_dict, variable_name='y')

    #Now get the various variables etc.
    if graphical_dimensionality == 2:
        independent_variables_dict, constants_extracted_dict, equation_extracted_dict, x_variable_extracted_dict, y_variable_extracted_dict = parse_equation_dict(equation_dict=equation_dict)
        constants_extracted_dict, equation_extracted_dict #These will not be used. The rest of this comment is to avoid a vs code pylint flag. # pylint: disable=unused-variable, disable=pointless-statement
    elif graphical_dimensionality == 3:
        independent_variables_dict, constants_extracted_dict, equation_extracted_dict, x_variable_extracted_dict, y_variable_extracted_dict, z_variable_extracted_dict = parse_equation_dict(equation_dict=equation_dict)
        constants_extracted_dict, equation_extracted_dict #These will not be used. The rest of this comment is to avoid a vs code pylint flag. # pylint: disable=unused-variable, disable=pointless-statement
    else:
        raise ValueError("Error: graphical_dimensionality not received and/or not evaluatable by current code.")

    #Start of block to check for any custom units and add them to the ureg if necessary.
    custom_units_list = []
    #helper function to clean custom units brackets. In future, could be made more general rather than hardcoded as angle brackets.
    def clean_brackets(string):
        return string.replace("<", "").replace(">", "")

    for constant_entry_key in independent_variables_dict.keys():
        independent_variables_string = independent_variables_dict[constant_entry_key]
        custom_units_extracted = extract_tagged_strings(independent_variables_string)
        independent_variables_dict[constant_entry_key] = clean_brackets(independent_variables_dict[constant_entry_key])
        for custom_unit in custom_units_extracted: #this will be skipped if the list is empty.
            ureg.define(f"{custom_unit} = [custom]") #use "[custom]" to create a custom unit in the pint module.
        custom_units_list.extend(custom_units_extracted)
      
    #now also check for the x_variable_extracted_dict 
    custom_units_extracted = extract_tagged_strings(x_variable_extracted_dict["units"])
    x_variable_extracted_dict["units"] = clean_brackets(x_variable_extracted_dict["units"])
    for custom_unit in custom_units_extracted: #this will be skipped if the list is empty.
        ureg.define(f"{custom_unit} = [custom]") #use "[custom]" to create a custom unit in the pint module.
    custom_units_list.extend(custom_units_extracted)

    #now also check for the y_variable_extracted_dict (technically not needed)
    custom_units_extracted = extract_tagged_strings(y_variable_extracted_dict["units"])
    y_variable_extracted_dict["units"] = clean_brackets(y_variable_extracted_dict["units"])
    for custom_unit in custom_units_extracted: #this will be skipped if the list is empty.
        ureg.define(f"{custom_unit} = [custom]") #use "[custom]" to create a custom unit in the pint module.
    custom_units_list.extend(custom_units_extracted)

    if graphical_dimensionality == 3:
        #now also check for the z_variable_extracted_dict (technically not needed)
        custom_units_extracted = extract_tagged_strings(z_variable_extracted_dict["units"])
        z_variable_extracted_dict["units"] = clean_brackets(z_variable_extracted_dict["units"])
        for custom_unit in custom_units_extracted: #this will be skipped if the list is empty.
            ureg.define(f"{custom_unit} = [custom]") #use "[custom]" to create a custom unit in the pint module.
        custom_units_list.extend(custom_units_extracted)

    #now also check for the equation_string
    custom_units_extracted = extract_tagged_strings(equation_string)
    equation_string = clean_brackets(equation_string)
    for custom_unit in custom_units_extracted: #this will be skipped if the list is empty.
        ureg.define(f"{custom_unit} = [custom]") #use "[custom]" to create a custom unit in the pint module.
    custom_units_list.extend(custom_units_extracted)        
    
    # Remove duplicates by converting to a set and back to a list.
    custom_units_list = list(set(custom_units_list))
    #now sort from longest to shortest, since we will have to put them back in that way later.
    # Sort the unique units by length in descending order
    custom_units_list = sorted(custom_units_list, key=len, reverse=True)

    #End of block to check for any custom units and add them to the ureg if necessary.

    #For graphical_dimensionality of 2, The full list of independent variables includes the x_variable and the independent_variables. 
    independent_variables = list(independent_variables_dict.keys())#.append(x_variable_extracted_dict['label'])
    independent_variables.append(x_variable_extracted_dict['label'])
    if graphical_dimensionality == 3: #for graphical_dimensionality of 3, the y_variable is also an independent variable.
        independent_variables.append(y_variable_extracted_dict['label'])

    #Now define the dependent variable:
    if graphical_dimensionality == 2:
        dependent_variable = y_variable_extracted_dict["label"]
    elif graphical_dimensionality == 3:
        dependent_variable = z_variable_extracted_dict["label"]
    else:
        raise ValueError("Error: graphical_dimensionality not received and/or not evaluatable by current code.")
    solved_coordinates_list = [] #These are x,y pairs or x,y,z triplets. can't just keep y_points, because there could be more than one solution.
    y_units = ''#just initializing.
    dependent_variable_units = '' #just initializing.

    if graphical_dimensionality == 2:
        input_points_list = x_points #currently a list of points [1,2,3]
        #nested_x_points = [[x] for x in input_points_list] #this way could have  [ [x1],[x2],...]
    elif graphical_dimensionality == 3:
        import itertools
        input_points_list = list(itertools.product(x_points, y_points))  #[ [x1,y1], [x1,y2] ]        
    else:
        raise ValueError("Error: graphical_dimensionality not received and/or not evaluatable by current code.")

    for current_point in input_points_list:
        #For each point, need to call the "solve_equation" equation (or a vectorized version of it).
        #This is the form that the variables need to take
        # # Example usage
        # independent_variables_values_and_units = {
        #     "x": "2 m / s",
        #     "y": "3 meter"
        # }
        # We also need to define the independent variables and dependent variables.
        if graphical_dimensionality == 2:
            independent_variables_dict[x_variable_extracted_dict["label"]] = str(current_point) + " " + x_variable_extracted_dict["units"]
        if graphical_dimensionality == 3:
            independent_variables_dict[x_variable_extracted_dict["label"]] = str(current_point[0]) + " " + x_variable_extracted_dict["units"]
            independent_variables_dict[y_variable_extracted_dict["label"]] = str(current_point[1]) + " " + y_variable_extracted_dict["units"]
        #if graphical_dimensionality is 2D, dependent_variable_solutions is y_solutions. 
        #if graphical_dimensionality is 3D, dependent_variable_solutions is z_solutions. 
        if verbose: print("json_equationer > equation_evaluator > evaluate_equation_dict > current_point:", current_point)
        dependent_variable_solutions = solve_equation(equation_string, independent_variables_values_and_units=independent_variables_dict, dependent_variable=dependent_variable)
        if dependent_variable_solutions:
            for dependent_variable_point_with_units in dependent_variable_solutions:
                if graphical_dimensionality == 2:
                    y_point = float(dependent_variable_point_with_units.split(" ", 1)[0]) #the 1 splits only at first space.
                    solved_coordinates_list.append([current_point, y_point])
                    if dependent_variable_units == '': #only extract units the first time.
                        y_units = dependent_variable_point_with_units.split(" ", 1)[1] #the 1 splits only at first space.
                if graphical_dimensionality == 3:
                    z_point = float(dependent_variable_point_with_units.split(" ", 1)[0]) #the 1 splits only at first space.
                    solved_coordinates_list.append([current_point[0],current_point[1], z_point])
                    if dependent_variable_units == '': #only extract units the first time.
                        z_units = dependent_variable_point_with_units.split(" ", 1)[1] #the 1 splits only at first space.
                
    #now need to convert the x_y_pairs.
    # Separating x and y points
    if graphical_dimensionality == 2:
        x_points, y_points = zip(*solved_coordinates_list)
    elif graphical_dimensionality == 3:
        x_points, y_points, z_points = zip(*solved_coordinates_list)

    # Convert tuples to lists
    x_points = list(x_points)
    y_points = list(y_points)
    if graphical_dimensionality == 3:
        z_points = list(z_points)
  
    #Some lines to ensure units are appropriate format before doing any inverse units conversions.
    if graphical_dimensionality == 2:
        x_units = x_variable_extracted_dict["units"]
        if "(" not in x_units:
            x_units = "(" + x_units + ")"
        if "(" not in y_units:
            y_units = "(" + y_units + ")"

    if graphical_dimensionality == 3:
        x_units = x_variable_extracted_dict["units"]
        y_units = y_variable_extracted_dict["units"]
        if "(" not in x_units:
            x_units = "(" + x_units + ")"
        if "(" not in y_units:
            y_units = "(" + y_units + ")"
        if "(" not in z_units:
            z_units = "(" + z_units + ")"

    y_units = convert_inverse_units(y_units)
    x_units = convert_inverse_units(x_units)
    if graphical_dimensionality == 3:
        z_units = convert_inverse_units(z_units)

    #Put back any custom units tags, only needed for dependent variable.
    if graphical_dimensionality == 2:
        y_units = return_custom_units_markup(y_units, custom_units_list)
    if graphical_dimensionality == 3:
        z_units = return_custom_units_markup(z_units, custom_units_list)

    #Fill the dictionary that will be returned.
    evaluated_dict = {}
    evaluated_dict['graphical_dimensionality'] = graphical_dimensionality
    evaluated_dict['x_units'] = x_units
    evaluated_dict['y_units'] = y_units
    evaluated_dict['x_points'] = x_points
    evaluated_dict['y_points'] = y_points
    if graphical_dimensionality == 3:
        z_units = return_custom_units_markup(z_units, custom_units_list)
        evaluated_dict['z_units'] = z_units
        evaluated_dict['z_points'] = z_points
    if graphical_dimensionality_added == True: #undo adding graphical_dimensionality if it was added by this function.
        equation_dict.pop("graphical_dimensionality")
    return evaluated_dict

if __name__ == "__main__":
    #Here is a 2D example:
    example_equation_dict = {
        'equation_string': 'k = A*(e**((-Ea)/(R*T)))',
        'x_variable': 'T (K)',  
        'y_variable': 'k (s**(-1))',
        'constants': {'Ea': '30000 (J)*(mol^(-1))', 'R': '8.314 (J)*(mol^(-1))*(K^(-1))' , 'A': '1*10^13 (s^-1)', 'e': '2.71828'},
        'num_of_points': 10,
        'x_range_default': [200, 500],
        'x_range_limits' : [],
        'x_points_specified' : [],
        'points_spacing': 'Linear',
        'reverse_scaling' : False
    }

    example_evaluated_dict = evaluate_equation_dict(example_equation_dict)
    print(example_evaluated_dict)

    #Here is a 3D example.
    example_equation_dict = {
        'equation_string': 'k = A*(e**((-Ea)/(R*T)))',
        'graphical_dimensionality' : 3,
        'x_variable': 'T (K)',  
        'y_variable': 'Ea (J)*(mol^(-1))',
        'z_variable': 'k (s**(-1))', 
        'constants': {'R': '8.314 (J)*(mol^(-1))*(K^(-1))' , 'A': '1*10^13 (s^-1)', 'e': '2.71828'},
        'num_of_points': 10,
        'x_range_default': [200, 500],
        'x_range_limits' : [],
        'y_range_default': [30000, 50000],
        'y_range_limits' : [],
        'x_points_specified' : [],
        'points_spacing': 'Linear',
        'reverse_scaling' : False
    }

    example_evaluated_dict = evaluate_equation_dict(example_equation_dict)
    print(example_evaluated_dict)




import os
import tkinter as tk
from tkinter import filedialog
from tkinterdnd2 import DND_FILES, TkinterDnD


#The below class creates a window for dragging and dropping or browsing and selecting files
#And each time one or more file is added, the full file list and most recently added files will be passed to
#The function supplied by the user (function_for_after_file_addition)
#with the two variables passed being all_selected_file_paths, newly_added_file_paths
#This class **cannot** be initiated directly, it should initiated using the
#companion function create_and_launch
class DragDropApp:
    def __init__(self, root, app_name = '', function_for_after_file_addition = None):
        self.root = root
        self.root.title(app_name)
        self.function_for_after_file_addition = function_for_after_file_addition

        # Enable native drag-and-drop capability
        self.root.drop_target_register(DND_FILES)
        self.root.dnd_bind("<<Drop>>", self.drop_files)

        # Create a drop zone
        self.drop_frame = tk.Label(root, text="Drag and drop files here \n\n Click End When Finished", bg="lightgray", width=50, height=10)
        self.drop_frame.pack(pady=10)

        # Create a listbox to display selected files
        self.file_listbox = tk.Listbox(root, width=60, height=10)
        self.file_listbox.pack(pady=10)

        # Buttons for manual selection and finalizing selection
        self.select_button = tk.Button(root, text="Select Files By Browsing", command=self.open_file_dialog)
        self.select_button.pack(pady=5)

        # Create a frame for the middle buttons
        button_frame_middle = tk.Frame(root)
        button_frame_middle.pack(pady=5)

        self.clear_button = tk.Button(button_frame_middle, text="Clear Files List", command=self.clear_file_list)  # New "Clear" button
        self.clear_button.pack(side = tk.LEFT, pady=5)

        # "Download Output" button
        self.download_button = tk.Button(button_frame_middle, text="Download Output", command=self.download_output)
        self.download_button.pack(side = tk.RIGHT, pady=5)

        self.done_button = tk.Button(root, text="End", command=self.finish_selection)
        self.done_button.pack(pady=5)

        # Store selected file paths
        self.all_selected_file_paths = []

    def clear_file_list(self):
        """Clears the listbox and resets selected files."""
        self.file_listbox.delete(0, tk.END)  # Clear listbox
        self.all_selected_file_paths = []  # Reset file list
        self.function_for_after_file_addition(all_selected_file_paths=[], newly_added_file_paths=[])
        print("File list cleared!")  # Optional debug message

    def open_file_dialog(self):
        """Opens a file dialog to manually select files."""
        newly_added_file_paths = self.root.tk.splitlist(tk.filedialog.askopenfilenames(title="Select files"))
        if newly_added_file_paths:
            self.all_selected_file_paths.extend(newly_added_file_paths)
            self.update_file_list(newly_added_file_paths)

    def drop_files(self, event):
        """Handles dropped files into the window."""
        newly_added_file_paths = self.root.tk.splitlist(event.data)
        if newly_added_file_paths:
            self.all_selected_file_paths.extend(newly_added_file_paths)
            self.update_file_list(newly_added_file_paths)

    def update_file_list(self, newly_added_file_paths):
        """Updates the listbox with selected filenames."""
        self.file_listbox.delete(0, tk.END)  # Clear listbox
        for filename_and_path in self.all_selected_file_paths:
            self.file_listbox.insert(tk.END, os.path.basename(filename_and_path))  # Show filenames only
        # If there is a function_for_after_file_addition, pass full list and newly added files into function_for_after_file_addition
        if self.function_for_after_file_addition is not None:
            output = self.function_for_after_file_addition(self.all_selected_file_paths, newly_added_file_paths)
            self.output_for_download = output[0] #store the first part of the output for download.

    def download_output(self):
        """Allows user to choose where to save the output."""
        if hasattr(self, "output_for_download"):
            file_path = filedialog.asksaveasfilename(filetypes=[("*.*", "*.txt")], title="Save Output As")
            if file_path:  # If a valid path is chosen
                with open(file_path, "w") as file:
                    file.write(str(self.output_for_download))
                print(f"Output saved as '{file_path}'!")
            else:
                print("File save operation canceled.")
        else:
            print("No output available to download.")


    def finish_selection(self):
        """Closes the window and returns selected files."""
        self.root.quit() # Close the window

# This function is a companion function to
# The class DragDropApp for creating a file selection and function call app
# The function_for_after_file_addition should return a list where the first item is something that can be downloaded.
def create_and_launch(app_name = '', function_for_after_file_addition=None):
    """Starts the GUI and returns selected files."""
    root = TkinterDnD.Tk()
    app = DragDropApp(root, app_name=app_name, function_for_after_file_addition=function_for_after_file_addition)
    root.mainloop() # Runs the Tkinter event loop
    return app.all_selected_file_paths # Returns selected files after the window closes
